#include <QPainter>
#include <QDebug>
#include "cgraphics2d.h"

QRect   CGraphics2D::NoRect;
QColor  CGraphics2D::Transparent;

CGraphics2D  graph2D;

//---------------------------------------------------------------------------------------------
CGraphics2D::CGraphics2D()
{
    NoRect.setWidth(-1);
    Transparent.setAlpha(0);
}

//---------------------------------------------------------------------------------------------
void    CGraphics2D::line(QImage* image, const CVec2& p1, const CVec2& p2, QColor color)
{
    if (image->valid(p1.x, p1.y) == false ||
        image->valid(p2.x, p2.y) == false)
        return;

    int32_t min, max;

    if (p1.y == p2.y)
    {
        // horizontal
        min = MIN(p1.x, p2.x);
        max = MAX(p1.x, p2.x);

        for (int32_t i = min; i <= max; ++i)
            image->setPixel(i, p1.y, color.rgba());
    }
    else
    {
        // vertical
        min = MIN(p1.y, p2.y);
        max = MAX(p1.y, p2.y);

        for (int32_t i = min; i <= max; ++i)
            image->setPixel(p1.x, i, color.rgba());
    }
}

//---------------------------------------------------------------------------------------------
void    CGraphics2D::rect(QImage* image, const CRect& r, QColor color, bool filled)
{
    if (image->valid(r.tl.x, r.tl.y) == false ||
        image->valid(r.br.x, r.br.y) == false)
        return;

    if (filled == false)
    {
        line(image, r.tl, CVec2(r.br.x, r.tl.y), color);
        line(image, CVec2(r.br.x, r.tl.y), r.br, color);
        line(image, r.tl, CVec2(r.tl.x, r.br.y), color);
        line(image, CVec2(r.tl.x, r.br.y), r.br, color);
    }
    else
    {
        int32_t min, max;

        min = MIN(r.tl.y, r.br.y);
        max = MAX(r.tl.y, r.br.y);

        for (int32_t i = min; i <= max; ++i)
            line(image, CVec2(r.tl.x, i), CVec2(r.br.x, i), color);
    }
}

//---------------------------------------------------------------------------------------------
void    CGraphics2D::drawImage(QImage* destImage, const CVec2& pos, const QImage& srcImage, int angle, bool flipX, QRect clipRect)
{
    QPainter painter(destImage);
    QPoint  pos2(pos.x, pos.y);

    if (clipRect.width() >= 0)
        painter.setClipRect(clipRect);

    if (angle == 0)
    {
        if (flipX == true)
        {
            painter.translate(pos2);
            painter.translate((float)srcImage.width() / 2, 0);
            painter.scale(-1, 1);
            painter.translate(-(float)srcImage.width() / 2, 0);
            painter.drawImage(0, 0, srcImage);
        }
        else
        {
            painter.drawImage(pos2, srcImage);
        }
        painter.resetTransform();
    }
    else
    {
        painter.translate(pos2);
        painter.translate((float)srcImage.width() / 2, (float)srcImage.height() / 2);
        painter.rotate(angle);
        painter.translate(-(float)srcImage.width() / 2, -(float)srcImage.height() / 2);
        painter.drawImage(0.5, 0.5, srcImage);
        painter.resetTransform();
    }
    painter.setClipping(false);

    painter.end();
}

//---------------------------------------------------------------------------------------------
void    CGraphics2D::drawImageScaled(QImage* destImage, const CVec2& pos, const QImage& srcImage, float scale, bool flipX, float scaleY)
{
    QPainter painter(destImage);

    painter.setRenderHint(QPainter::SmoothPixmapTransform, true);

    QTransform  transform;

    transform.translate(pos.x, pos.y);

    if (scaleY == 0.0f)
        transform.scale(scale, scale);
    else
        transform.scale(scale, scaleY);

    if (flipX == true)
    {
        transform.translate((float)srcImage.size().width() / 2.0f, 0);
        transform.scale(-1, 1);
        transform.translate(-(float)srcImage.size().width() / 2.0f, 0);
    }

    painter.setTransform(transform);

    painter.drawImage(0, 0, srcImage);

    painter.resetTransform();
    painter.end();
}

//---------------------------------------------------------------------------------------------
QImage CGraphics2D::cropImage(const QImage& srcImage, CRect rect)
{
    QRect   qrect(rect.tl.x, rect.tl.y, rect.br.x - rect.tl.x + 1, rect.br.y - rect.tl.y + 1);
    return srcImage.copy(qrect);
}

//---------------------------------------------------------------------------------------------
void    CGraphics2D::drawImageAtlas(QImage* destImage, const CVec2& pos, const QImage& srcImage, CRect srcRect, QColor color)
{
    QPainter painter(destImage);

    CVec2   rectSize = srcRect.br - srcRect.tl + CVec2(1, 1);
    QRect   srcQRect(srcRect.tl.x, srcRect.tl.y, rectSize.x, rectSize.y);

    if (color.alpha() == 0)
    {
        QRect   destQRect(pos.x, pos.y, rectSize.x, rectSize.y);
        painter.drawImage(destQRect, srcImage, srcQRect);
        painter.end();
    }
    else
    {
        QImage  tempImage(rectSize.x, rectSize.y, QImage::Format_ARGB32);
        tempImage.fill(color);

        QPainter painter2(&tempImage);
        painter2.setCompositionMode(QPainter::CompositionMode_DestinationIn);
        painter2.drawImage(QRect(0, 0, rectSize.x, rectSize.y), srcImage, srcQRect);
        painter2.setCompositionMode(QPainter::CompositionMode_SourceOver);
        painter2.end();

        painter.drawImage(pos.x, pos.y, tempImage);
        painter.end();
    }
}

//---------------------------------------------------------------------------------------------
void    CGraphics2D::darken(QImage* image, float opacity)
{
    QImage shadowImage(image->width(), image->height(), QImage::Format_ARGB32);
    shadowImage.fill(QColor(0, 0, 0));

    QPainter painter(image);
    painter.setOpacity(opacity);
    painter.setCompositionMode(QPainter::CompositionMode_SourceAtop);
    painter.drawImage(QPoint(0, 0), shadowImage);
    painter.setCompositionMode(QPainter::CompositionMode_SourceOver);
    painter.setOpacity(1.0f);
    painter.end();
}

//------------------------------------------------------------
void    CGraphics2D::patternRectangle(QImage* image, CRect rect)
{
    for (int y = rect.tl.y; y <= rect.br.y; ++y)
        for (int x = rect.tl.x; x <= rect.br.x; ++x)
            if (((x + y) & 1) == 0)
                image->setPixel(QPoint(x, y), 0);
}

//---------------------------------------------------------------------------------------------
void    CGraphics2D::Xor(QImage* image, CRect rect, QColor color)
{
    for (int y = rect.tl.y; y <= rect.br.y; ++y)
        for (int x = rect.tl.x; x <= rect.br.x; ++x)
        {
            QColor  newColor;
            QRgb pixel = image->pixel(QPoint(x, y));
            newColor.setRed(qRed(pixel) ^ color.red());
            newColor.setGreen(qGreen(pixel) ^ color.green());
            newColor.setBlue(qBlue(pixel) ^ color.blue());
            image->setPixel(QPoint(x, y), newColor.rgba());
        }
}
