#include "game.h"
#include "tiles.h"
#include "walls.h"
#include "doors.h"
#include "player.h"
#include "../common/sources/system/fileCache.h"
#include "../common/sources/system/cgraphics2d.h"
#include <QDebug>
#include <QPainter>
#include "interface.h"
#include "bridge.h"
#include "keyboard.h"
#include "objects.h"

CGame   game;

//-----------------------------------------------------------------------------------------
CGame::CGame()
{
    currentLevel = 0;
}

//-----------------------------------------------------------------------------------------
void CGame::drawWall(QImage* image, CVec2 mapPos, CVec2 tablePos, EWallSide side, bool flip)
{
    CTile*  tile = map.getTile(mapPos);
    EWallSide playerSide = player.getWallSide(side);

    if (tile->mWalls[playerSide].getType() != eWallNothing)
    {
        const char* fileName = walls.getFilename(tablePos, side, flip);

        if (fileName[0] != 0)
        {
            // draw the wall
            QImage  wallGfx = fileCache.getImage(fileName);
            CVec2 gfxPos = walls.getGfxPos(tablePos, side, flip);

            graph2D.drawImage(image, gfxPos, wallGfx, 0, flip);

            // draw the ornates
            CWall*  wall = &tile->mWalls[playerSide];

            if (wall->getType() == eWallSimple)
            {
                //------------------------------------------------------------------------------
                // simple wall
                int ornate = wall->getOrnateParam("Ornate");
                walls.drawOrnate(image, tablePos, side, ornate);
            }
            else if (wall->getType() == eWallPortrait)
            {
                //------------------------------------------------------------------------------
                // champion mirror
                walls.drawOrnate(image, tablePos, side, ORNATE_MIRROR);

                if (tablePos == CVec2(2, 3) && side == eWallSideUp)
                {
                    // if the mirror is right in front of us, we draw the champion's head
                    int champion = wall->getIntParam("Champion");
                    bool isEmpty = wall->getBoolParam("isEmpty");

                    if (isEmpty == false)
                    {
                        CRect   rect = interface.getChampionPortraitRect(champion);
                        QImage  portraits = fileCache.getImage("gfx/interface/ChampPortraits.png");
                        CVec2   pos(96, 68);
                        graph2D.drawImageAtlas(image, pos, portraits, rect);

                        int lastChampChosed;
                        for (lastChampChosed = 0; lastChampChosed < 4; ++lastChampChosed)
                            if (characters[lastChampChosed].portrait == -1)
                                break;

                        if (lastChampChosed != 4)
                        {
                            CRect   mouseRect(pos, CVec2(pos.x + CHAMP_PORTRAIT_WIDTH - 1, pos.y + CHAMP_PORTRAIT_HEIGHT - 1));
                            mouse.addArea(eMouseAreaG_Champion, mouseRect, eCursor_Hand, (void*)wall);
                        }
                    }
                }
            }
            else if (wall->getType() == eWallDoorSide)
            {
                //------------------------------------------------------------------------------
                // side of a door
                if (tablePos == CVec2(2, 3) && side == eWallSideUp)
                {
                    QImage  doorFrame = fileCache.getImage("gfx/3DView/doors/Frame_Face_2.png");
                    graph2D.drawImage(image, CVec2(96, 33), doorFrame);
                }
            }
            else if (wall->getType() == eWallTorch)
            {
                //------------------------------------------------------------------------------
                // torch holder
                CObjectStack*   stack = map.findObjectsStack(mapPos, playerSide);
                CRect   rect = walls.drawOrnate(image, tablePos, side, (stack != NULL ? ORNATE_TORCH_FULL : ORNATE_TORCH_EMPTY));

                if (tablePos == CVec2(2, 3) && side == eWallSideUp)
                    mouse.addArea(eMouseAreaG_WallTorch, rect, eCursor_Hand, (void*)mapPos.x, (void*)mapPos.y, (void*)playerSide);
            }
            else if (wall->getType() == eWallAlcove)
            {
                //------------------------------------------------------------------------------
                // alcove
                int type = wall->getEnumParam("Type");
                walls.drawOrnate(image, tablePos, side, ORNATE_ARCHED_ALCOVE + type);

                if (tablePos.y > 0 && side == eWallSideUp)
                {
                    objects.drawObjectsStack(image, mapPos, playerSide, tablePos);

                    if (tablePos == CVec2(2, 3) && mouse.mObjectInHand.getType() != 0)
                    {
                        CRect   rectLeft(CVec2(66, 84), CVec2(157, 124));
                        mouse.addArea(eMouseAreaG_DropObject, rectLeft, eCursor_Hand, (void*)mapPos.x, (void*)mapPos.y, (void*)playerSide);
                    }
                }
            }
            else if (wall->getType() == eWallText)
            {
                //------------------------------------------------------------------------------
                // wall text
                std::string text = wall->getStringParam("Text");
                text = interface.getText(text);

                if (tablePos == CVec2(2, 3) && side == eWallSideUp)
                {
                    QImage  patch = fileCache.getImage("gfx/3DView/TextPatch.png");
                    graph2D.drawImage(image, CVec2(111, 97), patch);

                    interface.drawTextMultiLines(image, CVec2(112, 74), CInterface::eFontWalls, text, QColor(), 0, 11);
                }
                else
                {
                    std::vector<std::string>    lines = interface.getTextLines(text);
                    walls.drawOrnate(image, tablePos, side, ORNATE_TEXT, lines.size());
                }
            }
        }
    }
}

//-----------------------------------------------------------------------------------------
void CGame::displayMainView(QImage* image)
{
    if (interface.isInInventory() == false && interface.isInDiskDialog() == false)
    {
        CRect   mainRect(CVec2(0, 33), CVec2(MAINVIEW_WIDTH - 1, 33 + MAINVIEW_HEIGHT - 1));
        mouse.addArea(eMouseArea_None, mainRect, eCursor_Hand);

        // draw the floor and the ceiling
        QImage  floorGfx = fileCache.getImage("gfx/3DView/Floor.png");
        QImage  ceilingGfx = fileCache.getImage("gfx/3DView/Ceiling.png");

        bool flip = (player.pos.x + player.pos.y + player.dir) & 1;

        graph2D.drawImage(image, CVec2(0, 33), ceilingGfx, 0, flip);
        graph2D.drawImage(image, CVec2(0, 72), floorGfx, 0, flip);

        // draw the walls
        CVec2   tablePos;
        CVec2   localPos;
        CVec2   mapPos;

        for (int y = 0; y < WALL_TABLE_HEIGHT; ++y)
            for (int x = 0; x < WALL_TABLE_WIDTH; ++x)
            {
                tablePos = walls.getDrawSequence(CVec2(x, y));
                localPos = walls.getMapPos(tablePos);
                mapPos = player.getPosFromLocal(localPos);
                CTile*  tile = map.getTile(mapPos);

                if (tile != NULL)
                {
                    // draw the walls
                    drawWall(image, mapPos, tablePos, eWallSideUp, flip);
                    drawWall(image, mapPos, tablePos, eWallSideLeft, flip);
                    drawWall(image, mapPos, tablePos, eWallSideRight, flip);

                    // draw the stairs
                    if (tile->getType() == eTileStairs)
                        tiles.drawStairs(image, tile, tablePos);

                    // draw the floor ornates
                    if (tile->getType() == eTileGround)
                    {
                        // ground tile
                        int ornate = tile->getOrnateParam("Ornate");
                        tiles.drawFloorOrnate(image, tablePos, ornate);
                    }
                    else if (tile->getType() == eTilePressPlate)
                    {
                        // pressure plate
                        int ornate = tile->getEnumParam("Type");

                        if (ornate != 0)
                            tiles.drawFloorOrnate(image, tablePos, ornate + 6);
                    }

                    // draw objects back row
                    objects.drawBackRow(image, mapPos, tablePos);

                    // draw the doors
                    if (tile->getType() == eTileDoor)
                        doors.draw(image, tile, tablePos);

                    // draw objects front row
                    objects.drawFrontRow(image, mapPos, tablePos);

                    // draw ennemies
                }
            }
    }
}

//-----------------------------------------------------------------------------------------
QImage& CGame::mainLoop()
{
    static QImage image(SCREEN_WIDTH,
                        SCREEN_HEIGHT,
                        QImage::Format_RGBA8888);

    // special events
    if (player.shouldTakeStairs)
        player.takeStairs();

    if (interface.shouldLoad == true)
    {
        loadGame(interface.saveName.c_str());
        interface.shouldLoad = false;
        interface.setMainState(CInterface::eMainLoadedDialog);
    }

    if (interface.shouldSave == true)
    {
        saveGame(interface.saveName.c_str());
        interface.shouldSave = false;
        interface.setMainState(CInterface::eMainSavedDialog);
    }

    // clear the mouse areas
    mouse.clearAreas();

    // clear the screen
    image.fill(BLACK);

    // display
    displayMainView(&image);
    interface.display(&image);

    // processing
    mouse.update();
    SMouseArea* clickedArea = mouse.clickedArea(&image);

    keyboard.update();
    update(clickedArea);
    interface.update(clickedArea);

    return image;
}

//-----------------------------------------------------------------------------------------
void CGame::update(SMouseArea* clickedArea)
{
    doors.update();
    tiles.updatePressPlates();

    if (clickedArea != NULL)
    {
        if (mouse.mButtonPressing == true)
        {
            if (clickedArea->type == eMouseAreaG_Champion)
            {
                lastMirrorClicked = (CWall*)clickedArea->param1;

                int lastChampChosed;
                for (lastChampChosed = 0; lastChampChosed < 4; ++lastChampChosed)
                    if (characters[lastChampChosed].portrait == -1)
                        break;

                characters[lastChampChosed].fromDB(lastMirrorClicked->getIntParam("Champion"));
                interface.enterResurrect(lastChampChosed);
            }
            else if (clickedArea->type == eMouseAreaG_DoorButton)
            {
                CTile*  tile = (CTile*)clickedArea->param1;
                bool isOpened = tile->getBoolParam("isOpened");
                tile->setBoolParam("isOpened", !isOpened);
            }
            else if (clickedArea->type == eMouseAreaG_PickObject)
            {
                CObjectStack*   stack = (CObjectStack*)clickedArea->param1;
                int index = (int)clickedArea->param2;
                mouse.mObjectInHand = stack->getObject(index);
                stack->removeObject(index);

                if (stack->getSize() == 0)
                    map.removeObjectsStack(stack->mPos, stack->mSide);

                characters[interface.selectedChampion].updateLoad();
            }
            else if (clickedArea->type == eMouseAreaG_DropObject)
            {
                CVec2   pos = CVec2((int)clickedArea->param1, (int)clickedArea->param2);
                int side = (int)clickedArea->param3;
                CObjectStack*   stack = map.addObjectsStack(pos, (EWallSide)side);
                stack->addObject(mouse.mObjectInHand);
                mouse.mObjectInHand.setType(0);
                characters[interface.selectedChampion].updateLoad();
            }
            else if (clickedArea->type == eMouseAreaG_WallTorch)
            {
                CVec2   pos;
                EWallSide   side;

                pos.x = (int)clickedArea->param1;
                pos.y = (int)clickedArea->param2;
                side = (EWallSide)((int)clickedArea->param3);

                int type = mouse.mObjectInHand.getType();
                CObjectStack* stack = map.findObjectsStack(pos, side);

                if (type == 0 && stack != NULL)
                {
                    mouse.mObjectInHand = stack->getObject(0);
                    map.removeObjectsStack(pos, side);
                }
                else if (type == 3 && stack == NULL)
                {
                    stack = map.addObjectsStack(pos, side);
                    stack->addObject(mouse.mObjectInHand);
                    mouse.mObjectInHand.setType(0);
                }
                characters[interface.selectedChampion].updateLoad();
            }
        }
    }
}

//-----------------------------------------------------------------------------------------
void CGame::loadLevel(int level, CVec2 pos, int side)
{
    // snapshot the current level
    snapshotLevel();
    currentLevel = level;

    // search if the new level was snapshoted...
    for (size_t i = 0; i < mVisitedLevels.size(); ++i)
    {
        CSnapshot*  s = &mVisitedLevels[i];

        if (s->level == level)
        {
            map.freeMemory();
            map = s->map;
            tiles.setPressPlates(s->pressPlateStates);
            player.pos = pos;
            player.dir = side;
            return;
        }
    }

    // ...else load the new level
    static char fileName[256];
    sprintf(fileName, "maps/level%02d.map", level);
    map.load(fileName);
    player.pos = pos;
    player.dir = side;
    tiles.initMap();
    walls.init();
}

//-----------------------------------------------------------------------------------------
CSnapshot::CSnapshot()
{
    pressPlateStates = NULL;
}

//-----------------------------------------------------------------------------------------
CSnapshot::CSnapshot(const CSnapshot& rhs)
{
    pressPlateStates = NULL;
    *this = rhs;
}

//-----------------------------------------------------------------------------------------
CSnapshot::~CSnapshot()
{
    freePressPlates();
}

//-----------------------------------------------------------------------------------------
void CSnapshot::freePressPlates()
{
    if (pressPlateStates != NULL)
        delete[] pressPlateStates;
}

//-----------------------------------------------------------------------------------------
void CSnapshot::copyPressPlates(bool* src)
{
    freePressPlates();
    int mapSize = map.mSize.x * map.mSize.y;

    pressPlateStates = new bool[mapSize];
    for (int i = 0; i < mapSize; ++i)
        pressPlateStates[i] = src[i];
}

//-----------------------------------------------------------------------------------------
CSnapshot&  CSnapshot::operator=(const CSnapshot& rhs)
{
    level = rhs.level;
    map = rhs.map;
    copyPressPlates(rhs.pressPlateStates);
    return *this;
}

//-----------------------------------------------------------------------------------------
void CGame::snapshotLevel()
{
    if (currentLevel != 0)
    {
        // if there is an entry for this level, delete it
        std::vector<CSnapshot>::iterator it;

        for (it = mVisitedLevels.begin(); it != mVisitedLevels.end(); ++it)
            if (it->level == currentLevel)
            {
                mVisitedLevels.erase(it);
                break;
            }

        // add a new entry
        CSnapshot   s;
        s.level = currentLevel;
        s.map = map;
        s.copyPressPlates(tiles.getPressPlates());
        mVisitedLevels.push_back(s);
    }
}

//---------------------------------------------------------------------------------------------
void CGame::saveGame(const char *fileName)
{
    snapshotLevel();
    FILE*       handle = fopen(fileName, "wb");

    if (handle != NULL)
    {
        // save current level and player pos
        fwrite(&currentLevel, sizeof(currentLevel), 1, handle);
        fwrite(&player.pos.x, sizeof(player.pos.x), 1, handle);
        fwrite(&player.pos.y, sizeof(player.pos.y), 1, handle);
        fwrite(&player.dir, sizeof(player.dir), 1, handle);

        // save characters
        for (int i = 0; i < 4; ++i)
            game.characters[i].save(handle);

        // save maps' snapshots
        int numSnapshots = mVisitedLevels.size();
        fwrite(&numSnapshots, sizeof(numSnapshots), 1, handle);

        for (int i = 0; i < numSnapshots; ++i)
            mVisitedLevels[i].save(handle);

        fclose(handle);
    }
}

//---------------------------------------------------------------------------------------------
void CGame::loadGame(const char* fileName)
{
    FILE*       handle = fopen(fileName, "rb");

    if (handle != NULL)
    {
        // load current level and player pos in temporary variable as they will be set by loadLevel()
        int currentFileLevel;
        CVec2   filePlayerPos;
        uint8_t filePlayerDir;
        fread(&currentFileLevel, sizeof(currentFileLevel), 1, handle);
        fread(&filePlayerPos.x, sizeof(filePlayerPos.x), 1, handle);
        fread(&filePlayerPos.y, sizeof(filePlayerPos.y), 1, handle);
        fread(&filePlayerDir, sizeof(filePlayerDir), 1, handle);

        // load characters
        for (int i = 0; i < 4; ++i)
            game.characters[i].load(handle);

        // load maps' snapshots
        mVisitedLevels.clear();
        int numSnapshots;
        fread(&numSnapshots, sizeof(numSnapshots), 1, handle);
        mVisitedLevels.resize(numSnapshots);

        for (int i = 0; i < numSnapshots; ++i)
            mVisitedLevels[i].load(handle);

        fclose(handle);

        currentLevel = 0; // avoid snapshot
        loadLevel(currentFileLevel, filePlayerPos, filePlayerDir);
    }
}

//---------------------------------------------------------------------------------------------
void    CSnapshot::load(FILE* handle)
{
    fread(&level, sizeof(level), 1, handle);
    map.load(handle);

    uint8_t state;
    freePressPlates();
    int mapSize = map.mSize.x * map.mSize.y;
    pressPlateStates = new bool[mapSize];

    for (int i = 0; i < map.mSize.x * map.mSize.y; ++i)
    {
        fread(&state, sizeof(uint8_t), 1, handle);
        pressPlateStates[i] = state;
    }

}

//---------------------------------------------------------------------------------------------
void    CSnapshot::save(FILE* handle)
{
    fwrite(&level, sizeof(level), 1, handle);
    map.save(handle);

    uint8_t state;
    for (int i = 0; i < map.mSize.x * map.mSize.y; ++i)
    {
        state = pressPlateStates[i];
        fwrite(&state, sizeof(uint8_t), 1, handle);
    }
}
